/* This file is part of nanolisp, a very small Lisp system.

Copyright (C) 2007  Luca Saiu

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. */

%{
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdbool.h>
#include <string.h>
#include <assert.h>
#include "nanolisp.h"
#include "parser.h"

/* Yes, this is part of a non-reentrant parser. I know it's ugly. */
nanolisp_sexpression_t nanolisp_result_of_parsing;

/* Show a parse error: */
//int yyerror(char* message);
%}

%option yylineno

DIGIT   [0-9]
LOWER   [a-z]
UPPER   [A-Z]
LETTER  {LOWER}|{UPPER}
UNDERSCORE \_

COMMA     \,
SEMICOLON \;

SIGN       (\-|\+)

INTEGER    {SIGN}?{DIGIT}+

EXPONENT   (e|E){SIGN}?{INTEGER}
FLOAT      {SIGN}?(({DIGIT}+\.{DIGIT}*)|({DIGIT}*\.{DIGIT}+)){EXPONENT}?

/* We accept as an identifier pretty much everything not containing whitespaces;
   we also forbid backslashes, because allowing them might lead users to think
   they can use escapes. They can't. For more elaborate identifier names users
   should write identifiers as strings. Raw identifiers can't end with a colon
   because otherwise they could be mistaken for labels: */
IDENTIFIER [^ \n\t\r\(\)\'\`\@,]+

/*CESCAPE    (\\.)|(\\{DIGIT})|(\\{DIGIT}{DIGIT})|(\\{DIGIT}{DIGIT}{DIGIT})*/
CESCAPE    (\\[\"\'\\abtnrv])
STRING     \"([^\n\"]|{CESCAPE})*\"

EMBEDDED_CODE \{([^}]|(\\\}))*\}

WHITESPACE       (" "|\t|\n|\r|\v)+
LINECOMMENT ;[^\n\r]*(\n|\r)
/* An hack to recognize a line comment at the EOF without newline: */
LINECOMMENT_AT_EOF ;[^\n\r]*
COMMENT       {LINECOMMENT}|{LINECOMMENT_AT_EOF}
LABEL         {RAW_IDENTIFIER}:

%%

{WHITESPACE}   { /* Do nothing */ }
{COMMENT}      { /* Do nothing */ }
{FLOAT}       { return FLOAT; }
{INTEGER}     { return INTEGER; }
\#\(          { return OPENPAR_ARRAY; }
\(            { return OPENPAR; }
\)            { return CLOSEPAR; }
\.            { return DOT; }
\'            { return QUOTE; }
\`            { return QUASIQUOTE; }
\,\@          { return UNQUOTE_SPLICING; }
\,            { return UNQUOTE; }
\#t           { return TRUE_; }
\#f           { return FALSE_; }
{IDENTIFIER}  { return SYMBOL; }

%%

int yywrap(){
  return 1;
}

int yyerror(char* message){
  fprintf(stderr, "%s near \"%s\" at line %i\n", message, yytext, yylineno);
  nanolisp_error("parse error");
}

/* This is defined in the C source file generated by Bison: */
int yyparse();

/* Parse using the given buffer state, which must already be avtive, and
   in case of success write the result (as a nanolisp list) into the
   given pointer.
   Return 0 in case of success, and a nonzero value in case of errors. 
   Destroy the buffer_state in any case. */
static int nanolisp_parse_and_delete_buffer(YY_BUFFER_STATE buffer_state,
                                            nanolisp_sexpression_t *result){
  int result_of_yyparse = yyparse();
  yy_delete_buffer(buffer_state);
  if(result_of_yyparse == 0){
    *result = nanolisp_result_of_parsing;
    return 0;
  }
  else
    return -1;
}

/* Parse the given string (which should contain a sequence of s-expressions)
   and, in case of success, write the result (as a nanolisp list) into the
   given pointer.
   Return 0 in case of success, and a nonzero value in case of errors. */
int nanolisp_parse_string(const char *text, nanolisp_sexpression_t *result){
  return nanolisp_parse_and_delete_buffer(yy_scan_string(text), result);
}

/* Parse the content of the given FILE* (which should contain a sequence of
   s-expressions) and, in case of success, write the result (as a nanolisp
   list) into the given pointer.
   Return 0 in case of success, and a nonzero value in case of errors. */
int nanolisp_parse_FILE_star(FILE *file, nanolisp_sexpression_t *result){
  YY_BUFFER_STATE buffer_state = yy_create_buffer(file, YY_BUF_SIZE);
  yy_switch_to_buffer(buffer_state);
  return nanolisp_parse_and_delete_buffer(buffer_state, result);
}
